/*
*********************************************************************
*	
*	environment.c
*	Checks operating system environment
*		
*	Rainer Fuchs
*	EMBL Data Library
*	Postfach 10.2209
*	D-6900 Heidelberg, FRG
*	E-mail: fuchs@embl-heidelberg.de
*
*	Copyright  1992 EMBL Data Library
*		
**********************************************************************
*	
*/ 

#include <Traps.h>
#include <GestaltEqu.h>
#include <Folders.h>

#include "EMBL-Search.h"
#include "EMBL-Search.rsrc.h"

#include "environment.h"

#include "util.h"

/*
******************************* Prototypes ***************************
*/

static Boolean IsGestaltAvail(void);
static short GetSystemVersion(void);
static Boolean IsTempMemAvail(void);
static Boolean IsColorQDAvail(void);
static Boolean IsBalloonHelpAvail(void);
static Boolean IsAppleEventsAvail(void);
static Boolean IsFindFolderAvail(void);
static Boolean IsAliasMgrAvail(void);
static Boolean IsFSSpecAvail(void);
static Boolean IsFPUAvail(void);
static Boolean IsNewSFAvail(void);
static OSErr FindSysFolder(short *vRefNum, long *dirID);
static OSErr FindTempFolder(short *vRefNum, long *dirID);
static OSErr FindPrefFolder(short *vRefNum, long *dirID);
static Boolean TrapAvailable(short theTrap);
static TrapType GetTrapType(short theTrap);
static short NumToolboxTraps(void);
static long GetGestaltResult(OSType gestaltSelector);


/*
******************************** Global variables *****************
*/

SystemInfo gSystem;

/**************************************
*	Check system version, availability of certain functions, etc.
*/

void CheckEnvironment()
{
	Handle	apParam;
	
	/* Can we use Gestalt? */
	gSystem.bHasGestalt=IsGestaltAvail();

	/* No support for System versions less than 6.0.x */
	if ( (gSystem.sysVersion=GetSystemVersion()) < 0x0605 )
		FatalErrorMsg(ERR_VERSION);

	/* Get application name */
	GetAppParms(gSystem.appName, &gSystem.appResRef, &apParam);

	/* Check for availability of some specific functions */
	gSystem.bHasTempMemory		= IsTempMemAvail();
	gSystem.bHasBalloonHelp		= IsBalloonHelpAvail();
	gSystem.bHasFindFolder		= IsFindFolderAvail();
	gSystem.bHasAppleEvents		= IsAppleEventsAvail();
	gSystem.bHasAliasMgr			= IsAliasMgrAvail();
	gSystem.bHasFSSpec			= IsFSSpecAvail();
	gSystem.bHasColorQD			= IsColorQDAvail();
	gSystem.bHasFPU				= IsFPUAvail();
	gSystem.bHasNewSFCalls		= IsNewSFAvail();
	
	/* Get system folder vRefNum */
	FindSysFolder(&gSystem.sysVRefNum,&gSystem.sysDirID); /* error check ... */
	
	/* Find temporary items folder
		If FindFolder is available we use/create the "Temporary Items" folder,
		otherwise we simply use the System folder (we don't look for a
		subfolder in this case, because names are language-dependent
	*/
	
	FindTempFolder(&gSystem.tempVRefNum,&gSystem.tempDirID);
	
	/* Find preferences folder
		If FindFolder is available we use/create the "Preferences" folder,
		otherwise we simply use the System folder (we dont look for a
		subfolder in this case, because names are language-dependent
	*/
	FindPrefFolder(&gSystem.prefVRefNum,&gSystem.prefDirID);
}

/**************************************
*	Check for availability of Gestalt
*	Return value:	TRUE, if Gestalt available
*						FALSE, if not
*/

static Boolean IsGestaltAvail()
{
	return(TrapAvailable(_Gestalt));
}

/**************************************
*	Get System version
*	Return value:	System version, or 0 if not available
*/

static short GetSystemVersion()
{
	SysEnvRec theWorld;

	if(gSystem.bHasGestalt) {
		return( (short)GetGestaltResult (gestaltSystemVersion) );
	}
	else {
		return( (SysEnvirons(curSysEnvVers,&theWorld) == noErr) ? theWorld.systemVersion : 0);
	}
}

/**************************************
*	Check for availability of temporary memory calls
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsTempMemAvail()
{
	long result;
	
	result=GetGestaltResult(gestaltOSAttr);
	if(result && (result & (1 << gestaltTempMemSupport)))
		return(TRUE);
	else return(FALSE);

}

/**************************************
*	Check for availability of color QuickDraw
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsColorQDAvail()
{
	long result;
	SysEnvRec theWorld;
	
	if(gSystem.bHasGestalt) {
		result = GetGestaltResult(gestaltQuickdrawVersion);
		if(result && result >= gestalt8BitQD)
			return(TRUE);
		else return(FALSE);
	}
	else
		return( (SysEnvirons(curSysEnvVers,&theWorld) == noErr) ? theWorld.hasColorQD : FALSE);
}

/**************************************
*	Check for availability of FPU
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsFPUAvail()
{
	long result;
	SysEnvRec theWorld;
	
	if(gSystem.bHasGestalt) {
		result = GetGestaltResult(gestaltFPUType);
		if(result > gestaltNoFPU)
			return(TRUE);
		else return(FALSE);
	}
	else
		return( (SysEnvirons(curSysEnvVers,&theWorld) == noErr) ? theWorld.hasFPU : FALSE);
}

/**************************************
*	Check for Help Manager
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsBalloonHelpAvail()
{
	long result;
	
	result=GetGestaltResult(gestaltHelpMgrAttr);
	if(result && (result & (1 << gestaltHelpMgrPresent)))
		return(TRUE);
	else return(FALSE);

}

/**************************************
*	Check for availability of System 7 Standard File dialog calls
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsNewSFAvail()
{
	long result;
	
	result=GetGestaltResult(gestaltStandardFileAttr);
	if(result && (result & (1 << gestaltStandardFile58)))
		return(TRUE);
	else return(FALSE);

}

/**************************************
*	Check for AppleEvents
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsAppleEventsAvail()
{
	long result;
	
	result=GetGestaltResult(gestaltAppleEventsAttr);
	if(result && (result & (1 << gestaltAppleEventsPresent)))
		return(TRUE);
	else return(FALSE);

}

/**************************************
*	Check for FindFolder
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsFindFolderAvail()
{
	long result;
	
	result=GetGestaltResult(gestaltFindFolderAttr);
	if(result && (result & (1 << gestaltFindFolderPresent)))
		return(TRUE);
	else return(FALSE);

}

/**************************************
*	Check for Alias Manager
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsAliasMgrAvail()
{
	long result;
	
	result=GetGestaltResult(gestaltAliasMgrAttr);
	if(result && (result & (1 << gestaltAliasMgrPresent)))
		return(TRUE);
	else return(FALSE);

}

/**************************************
*	Check for System 7 File Manager calls
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean IsFSSpecAvail()
{
	long result;
	
	result=GetGestaltResult(gestaltFSAttr);
	if(result && (result & (1 << gestaltHasFSSpecCalls)))
		return(TRUE);
	else return(FALSE);

}

/**************************************
*	Obtain working directory refnum for system folder
*	Return value:	working directory refnum or 0, if not available
*/

static OSErr FindSysFolder(short *vRefNum, long *dirID)
{
	OSErr err;
	SysEnvRec	envRec;
	long procID;
	
	err=noErr;
	
	/* find folder */
	if(gSystem.bHasFindFolder)
		err=FindFolder(kOnSystemDisk, kSystemFolderType, kCreateFolder, 
				vRefNum, dirID);
				
	if(!gSystem.bHasFindFolder || err != noErr) {
		if ((err = SysEnvirons (curSysEnvVers, &envRec)) == noErr) {
			err = GetWDInfo(envRec.sysVRefNum,vRefNum,dirID,&procID);
		}
	}
	
	return(err);
}

/**************************************
*	Obtain working directory refnum for "Temporary Items" folder.
*	Return value:	working directory refnum or 0, if not available
*/

static OSErr FindTempFolder(short *vRefNum, long *dirID)
{
	OSErr err;
		
	err=noErr;
	
	/* find folder and create it if necessary */
	if(gSystem.bHasFindFolder)
		err=FindFolder(kOnSystemDisk, kTemporaryFolderType, kCreateFolder, 
				vRefNum, dirID);
				
	if(!gSystem.bHasFindFolder || err != noErr)
		err = FindSysFolder(vRefNum,dirID);
		
	return(err);
}


/**************************************
*	Get working directory refnum for "Preferences" folder.
*	Return value:	working directory refnum or 0, if not available
*/

static OSErr FindPrefFolder(short *vRefNum, long *dirID)
{
	OSErr err;
	
	err=noErr;
	
	/* find folder and create it if necessary */
	if(gSystem.bHasFindFolder)
	err=FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder, 
				vRefNum, dirID);
				
	if(!gSystem.bHasFindFolder || err != noErr)
		err = FindSysFolder(vRefNum,dirID);
		
	return(err);
}

/**************************************
*	Check availability of trap
*	Return value:	TRUE, if available
*						FALSE, if not
*/

static Boolean TrapAvailable(short theTrap)
{
	TrapType tType;
	
	tType=GetTrapType(theTrap);
	if(tType == ToolTrap) {
		theTrap &= 0x07FF;
		if(theTrap >= NumToolboxTraps())
			theTrap = _Unimplemented;
	}
	return( 	NGetTrapAddress(theTrap,tType) !=
				NGetTrapAddress(_Unimplemented,ToolTrap) );
}

/**************************************
*	Obtain trap type
*	Return value:	trap type
*/

static TrapType GetTrapType(short theTrap)
{
#define TrapMask 0x0800

	return( (theTrap & TrapMask > 0) ? ToolTrap : OSTrap);
}

/**************************************
*	Check size of trap table
*	Return value:	size of trap table
*/
		
static short NumToolboxTraps()
{
	return( (NGetTrapAddress(_InitGraf, ToolTrap) ==
				NGetTrapAddress(0xAA6E,ToolTrap) ) ? 0x200 : 0x400);
}

/**************************************
*	Call Gestalt to get system information.
*	Return value:	value returned by Gestalt, or 0 if Gestalt is not available
*/

static long	GetGestaltResult(OSType gestaltSelector)
{
	long	gestaltResult;
	
	if(!gSystem.bHasGestalt)
		return(0L);
		
	if (Gestalt(gestaltSelector, &gestaltResult) == noErr)
		return(gestaltResult);
	else
		return(0);
}

