/*
*********************************************************************
*	
*	pref.c
*	Preferences file
*		
*	Rainer Fuchs
*	EMBL Data Library
*	Postfach 10.2209
*	D-6900 Heidelberg, FRG
*	E-mail: fuchs@embl-heidelberg.de
*
*	Copyright  1992 EMBL Data Library
*	
*	The structure of the Prefs file is simple yet flexible. Each element of gPrefs has
*	got a "tag" to identify it. Written out to the Prefs file is the tag (0-255), two
*	bytes specifying the length of the following data, and the value of the gPrefs
*	element. Using this tag-value approach keeps the Prefs file upward and downward
*	compatible and extendible compared to a simple dump of gPrefs.
**********************************************************************
*	
*/

#include <stdio.h>

#include "EMBL-Search.h"
#include "EMBL-Search.rsrc.h"

#include "pref.h"
#include "tag.h"

#include "export.h"
#include "environment.h"
#include "pstr.h"
#include "util.h"

/*
******************************* Prototypes ****************************
*/


static OSErr ReadPrefs(void);
static void SetPref(Byte tag, Byte *data);
static void InitPrefs(void);
static Boolean Write1Pref(short output, PrefHeader *ph, Byte *data);



/*
********************************* Globals *****************************
*/

Prefs gPrefs;
OSType gCreatorSig[11] = {'????','MSWD','MACA','ttxt','MWII','nX^n','R*ch',
									'KAHL','TBB5','????','????'};

/**************************************
*	Read and set preferences
*/

void GetPrefs()
{
	OSErr err;
	
	InitPrefs();
	err=ReadPrefs();
	if(err != noErr) {		/* on error, we reset preferences */
		InitPrefs();
		if( err == fnfErr)	/* and create a new Prefs file if necessary */
			WritePrefs();
	}
}

/**************************************
*	Read Preferences file
*	Return value:	noErr, if successful
*						OS error code, if error occurred
*/

static OSErr ReadPrefs()
{
	OSErr		err;
	short		input;
	unsigned long		count;
	Str255	fName;
	PrefHeader ph;
	Byte		data[512];
	FSSpec	theFSS;
	
	/* get file name */
	GetIndString(fName,OTHERS,PREFSNAME);

	/* open file */
	MakeFSS(gSystem.prefVRefNum,fName,gSystem.prefDirID,&theFSS);
	err = OpenMacFileReadOnly(&theFSS,&input,FALSE);
	if(err != noErr) {
		if(err == fnfErr)
   		ErrorMsg(ERR_NOPREF);
   	else {
			sprintf(gError,LoadErrorStr(ERR_OPENPREF,FALSE),err);
			ErrorMsg(0);
		}
		return(err);
	}
	else {
		while(err == noErr) {	/* continue until we hit EOF */
			/* read header */
			count = sizeof(PrefHeader);
			err=ReadMacFile( input, (long *)&count, &ph,fName, FALSE );
			if(err == eofErr) {	/* hit end of file */
				err = noErr;
				break;
			}
			if(err == noErr) {
				/* read data */
    			if(ph.dataLen >512) ph.dataLen = 512; /* should issue a warning here */
    			count=(unsigned long)ph.dataLen;
    			err=ReadMacFile( input, (long *)&count, &data,fName, FALSE );
    		}
			if( err != noErr ) {
				sprintf(gError,LoadErrorStr(ERR_READPREF,FALSE),err);
				ErrorMsg(0);
			}
			else
				SetPref(ph.tag,data);
		}
		
   	FSClose(input);
   	return(err);
   }
}

/**************************************
*	Set a global preference defined by "tag" to the value given in "data"
*/

static void SetPref(Byte tag, Byte *data)
{
	switch(tag) {
		case T_EMBLINXDIR:
			gPrefs.inxDirSpec[DB_EMBL] = *(DirSpec *)data;
			break;
		case T_SWISSINXDIR:
			gPrefs.inxDirSpec[DB_SWISS] = *(DirSpec *)data;
			break;
		case T_PROSDATINXDIR:
			gPrefs.inxDirSpec[DB_PROSITEDAT] = *(DirSpec *)data;
			break;
		case T_PROSDOCINXDIR:
			gPrefs.inxDirSpec[DB_PROSITEDOC] = *(DirSpec *)data;
			break;
		case T_ENZYMEINXDIR:
			gPrefs.inxDirSpec[DB_ENZYME] = *(DirSpec *)data;
			break;
		case T_EPDINXDIR:
			gPrefs.inxDirSpec[DB_EPD] = *(DirSpec *)data;
			break;
/*****
		case T_SEQANALREFINXDIR:
			gPrefs.inxDirSpec[DB_SEQANALREF] = *(DirSpec *)data;
			break;

*****/		
		case T_EMBLDATADIR:
			gPrefs.dataDirSpec[DB_EMBL] = *(DirSpec *)data;
			break;
		case T_SWISSDATADIR:
			gPrefs.dataDirSpec[DB_SWISS] = *(DirSpec *)data;
			break;
		case T_PROSDATDATADIR:
			gPrefs.dataDirSpec[DB_PROSITEDAT] = *(DirSpec *)data;
			break;
		case T_PROSDOCDATADIR:
			gPrefs.dataDirSpec[DB_PROSITEDOC] = *(DirSpec *)data;
			break;
		case T_ENZYMEDATADIR:
			gPrefs.dataDirSpec[DB_ENZYME] = *(DirSpec *)data;
			break;
		case T_EPDDATADIR:
			gPrefs.dataDirSpec[DB_EPD] = *(DirSpec *)data;
			break;
/*****
		case T_SEQANALREFDATADIR:
			gPrefs.dataDirSpec[DB_SEQANALREF] = *(DirSpec *)data;
			break;

*****/
		case T_FORMAT:
			gPrefs.format = *(short *)data;
			break;
		case T_CREATOR:
			gPrefs.creator = *(short *)data;
			break;
		case T_CREATORNAME:
			pstrcpy(gPrefs.creatorName,(StringPtr)data);
			break;
		case T_CREATORSIG:
			gPrefs.creatorSig = *(OSType *)data;
			break;
		case T_CONFIRMCHG:
			gPrefs.confirmChg = *(Boolean *)data;
			break;
		case T_STARTQOPEN:
			gPrefs.startQOpen = *(Boolean *)data;
			break;
		case T_INCLUDELFCOPY:
			gPrefs.includeLFCopy = *(Boolean *)data;
			break;
		case T_MARKHYPERLINK:
			gPrefs.markHyperLink = *(Boolean *)data;
			break;
	}
}

/**************************************
*	Set default values for Preferences
*	Return value:	none
*/

static void InitPrefs()
{
	register short i;
	
	for(i=0;i<DB_NUM;++i) {
		*gPrefs.inxDirSpec[i].volName = EOS;
		gPrefs.inxDirSpec[i].dirID = 0;
		*gPrefs.dataDirSpec[i].volName = EOS;
		gPrefs.dataDirSpec[i].dirID = 0;
	}
	gPrefs.format=EMBL_FORMAT;
	gPrefs.creator=WORD_I;
	*gPrefs.creatorName = EOS;
	gPrefs.creatorSig = gCreatorSig[WORD_I];
	gPrefs.confirmChg = TRUE;
	gPrefs.startQOpen = TRUE;
	gPrefs.includeLFCopy = TRUE;
	gPrefs.markHyperLink = TRUE;
}

/**************************************
*	Write Preferences file
*/

void WritePrefs()
{
	OSErr		err;
	short		output;
	long		count;
	Str255	fName;
	PrefHeader ph;
	Byte		data[512];
	Boolean	ret;
	FSSpec	theFSS;
	
	/* get file name */	
	GetIndString(fName,OTHERS,PREFSNAME);
	
	/* create new version */
	MakeFSS(gSystem.prefVRefNum,fName,gSystem.prefDirID,&theFSS);
	if( (err=CreateMacFile(&theFSS,kApplSignature,kPrefFileType,FALSE)) != noErr) {
		sprintf(gError,LoadErrorStr(ERR_CREATPREF,FALSE),err);
		ErrorMsg(0);
	}
   else {	/* open file */
   	if( (err=OpenMacFile(&theFSS,&output,TRUE)) == noErr) {
   		/* write data */
   		
   		ph.tag = T_EMBLINXDIR;
   		ph.dataLen = (unsigned short) sizeof(DirSpec);
   		*(DirSpec *)data = gPrefs.inxDirSpec[DB_EMBL];
   		ret = Write1Pref(output,&ph,data);

			if(ret) {
				ph.tag = T_SWISSINXDIR;
				ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.inxDirSpec[DB_SWISS];
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_PROSDATINXDIR;
				ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.inxDirSpec[DB_PROSITEDAT];
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_PROSDOCINXDIR;
				ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.inxDirSpec[DB_PROSITEDOC];
   			ret = Write1Pref(output,&ph,data);
   		}
   		
   		if(ret) {
   			ph.tag = T_ENZYMEINXDIR;
   			ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.inxDirSpec[DB_ENZYME];
   			ret = Write1Pref(output,&ph,data);
   		}

   		if(ret) {
   			ph.tag = T_EPDINXDIR;
   			ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.inxDirSpec[DB_EPD];
   			ret = Write1Pref(output,&ph,data);
   		}
/*****

   		if(ret) {
   			ph.tag = T_SEQANALREFINXDIR;
   			ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.inxDirSpec[DB_SEQANALREF];
   			ret = Write1Pref(output,&ph,data);
   		}

*****/
			if(ret) {
				ph.tag = T_EMBLDATADIR;
				ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.dataDirSpec[DB_EMBL];
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_SWISSDATADIR;
				ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.dataDirSpec[DB_SWISS];
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_PROSDATDATADIR;
				ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.dataDirSpec[DB_PROSITEDAT];
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_PROSDOCDATADIR;
				ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.dataDirSpec[DB_PROSITEDOC];
   			ret = Write1Pref(output,&ph,data);
   		}
   		
   		if(ret) {
   			ph.tag = T_ENZYMEDATADIR;
   			ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.dataDirSpec[DB_ENZYME];
   			ret = Write1Pref(output,&ph,data);
   		}

   		if(ret) {
   			ph.tag = T_EPDDATADIR;
   			ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.dataDirSpec[DB_EPD];
   			ret = Write1Pref(output,&ph,data);
   		}

/*****
   		if(ret) {
   			ph.tag = T_SEQANALREFDATADIR;
   			ph.dataLen = (unsigned short) sizeof(DirSpec);
   			*(DirSpec *)data = gPrefs.dataDirSpec[DB_SEQANALREF];
   			ret = Write1Pref(output,&ph,data);
   		}

*****/
			if(ret) {
				ph.tag = T_FORMAT;
				ph.dataLen = (unsigned short) sizeof(short);
   			*(short *)data = gPrefs.format;
   			ret = Write1Pref(output,&ph,data);
   		}
   		
 			if(ret) {
				ph.tag = T_CREATOR;
				ph.dataLen = (unsigned short) sizeof(short);
   			*(short *)data = gPrefs.creator;
   			ret = Write1Pref(output,&ph,data);
   		}
   		
 			if(ret) {
				ph.tag = T_CREATORNAME;
				ph.dataLen = (unsigned short) pstrlen(gPrefs.creatorName) + 1;
   			pstrcpy((StringPtr)data,gPrefs.creatorName);
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_CREATORSIG;
				ph.dataLen = (unsigned short) sizeof(OSType);
   			*(OSType *)data = gPrefs.creatorSig;
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_CONFIRMCHG;
				ph.dataLen = (unsigned short) sizeof(Boolean);
   			*(Boolean *)data = gPrefs.confirmChg;
   			ret = Write1Pref(output,&ph,data);
   		}
   		
			if(ret) {
				ph.tag = T_STARTQOPEN;
				ph.dataLen = (unsigned short) sizeof(Boolean);
   			*(Boolean *)data = gPrefs.startQOpen;
   			ret = Write1Pref(output,&ph,data);
   		}

			if(ret) {
				ph.tag = T_INCLUDELFCOPY;
				ph.dataLen = (unsigned short) sizeof(Boolean);
   			*(Boolean *)data = gPrefs.includeLFCopy;
   			ret = Write1Pref(output,&ph,data);
   		}

			if(ret) {
				ph.tag = T_MARKHYPERLINK;
				ph.dataLen = (unsigned short) sizeof(Boolean);
   			*(Boolean *)data = gPrefs.markHyperLink;
   			ret = Write1Pref(output,&ph,data);
   		}

    		if(!ret) {
				sprintf(gError,LoadErrorStr(ERR_WRITEPREF,FALSE),err);
				ErrorMsg(0);
			}
			
   		FSClose(output);
		}
   }
}

/**************************************
*	Write header and data block for one Preferences field
*/

static Boolean Write1Pref(short output,PrefHeader *ph,Byte *data)
{
	OSErr				err;
	unsigned long	count;

	count = sizeof(PrefHeader);
	err = WriteMacFile( output, (long *)&count, ph,"\p",FALSE);
	if(err == noErr) {
	 	count = (unsigned long) ph->dataLen;
	 	err = WriteMacFile( output, (long *)&count, data,"\p",FALSE);
	}

	return(err == noErr);
}

/**************************************
*	Select a new text file creator
*	Return value:	True, if user selected one and no error occurred
*						False, if user canceled or an error occured
*/

Boolean PickNewCreator()
{
	DialogPtr	myDialog;
	Point			where;
	SFTypeList	myTypes;
	SFReply 		reply;
	StandardFileReply sys7reply;
	FInfo			fndrInfo;
	OSErr			err;
	
	/* show pick dialog */
	myDialog=GetNewDialog(PICKCREATOR_DLG,NULL,(WindowPtr)-1);
	
	if(gSystem.bHasNewSFCalls)
		CenterSFDlg(sfGetDialogID,&where);
	else
		CenterSFDlg(getDlgID,&where);
	MoveWindow(myDialog,where.h,where.v-39,TRUE);
	ShowWindow(myDialog);
	DrawDialog(myDialog);

	myTypes[0]='APPL';
	InitCursor();
	if(gSystem.bHasNewSFCalls) {
		/* we use Custom... in order to position the dialog */
		CustomGetFile(NULL,1,myTypes,&sys7reply,0,where,
							NULL,NULL,NULL,NULL,NULL);
   	DisposDialog(myDialog);
   	if(sys7reply.sfGood == TRUE) {
   		if((err = FSpGetFInfo(&sys7reply.sfFile,&fndrInfo)) != noErr) {
	  			sprintf(gError,LoadErrorStr(ERR_OPENFILE,FALSE),
	  						p2cstr(sys7reply.sfFile.name),err);
	   		c2pstr((char *)sys7reply.sfFile.name);
		  		return(ErrorMsg(0));
		  	}
	   	else {
	   		pstrcpy(gPrefs.creatorName,sys7reply.sfFile.name);
	   		gPrefs.creatorSig = fndrInfo.fdCreator;
	   	}
		}
		
		return(sys7reply.sfGood);   	
   }
	else {
		SFGetFile(where,"\p",NULL,1,myTypes,NULL,&reply); /* ask for file name */
		DisposDialog(myDialog);
   
	   if (reply.good) {							/* File was selected 					*/
	   	if((err = GetFInfo(reply.fName,reply.vRefNum,&fndrInfo)) != noErr) {
	  			sprintf(gError,LoadErrorStr(ERR_OPENFILE,FALSE),p2cstr(reply.fName),err);
	   		c2pstr((char *)reply.fName);
		  		return(ErrorMsg(0));
		  	}
	   	else {
	   		pstrcpy(gPrefs.creatorName,reply.fName);
	   		gPrefs.creatorSig = fndrInfo.fdCreator;
	   	}
		}
		
		return(reply.good);
	}
}

/**************************************
*	Show Options dialog and let user set some general preferences
*/

void GeneralOptions()
{
	DialogPtr	myDialog;
	Boolean		bQuit = FALSE,bChanged = FALSE;
	short			itemHit;
	
	myDialog=GetNewCenteredDialog(OPTIONS_DLG,66);

	/* show current settings */
	if(gPrefs.confirmChg)
		SetRadioButton(myDialog,CONFIRM_BOX,BTNON);
	else
		SetRadioButton(myDialog,CONFIRM_BOX,BTNOFF);
		
	if(gPrefs.startQOpen)
		SetRadioButton(myDialog,STARTQ_BOX,BTNON);
	else
		SetRadioButton(myDialog,STARTQ_BOX,BTNOFF);	
	
	if(gPrefs.includeLFCopy)
		SetRadioButton(myDialog,INCLUDELF_BOX,BTNON);
	else
		SetRadioButton(myDialog,INCLUDELF_BOX,BTNOFF);	

	if(gPrefs.markHyperLink)
		SetRadioButton(myDialog,HYPERLINK_BOX,BTNON);
	else
		SetRadioButton(myDialog,HYPERLINK_BOX,BTNOFF);	

	/* install user items to draw frame and default button outline */
	InstallUserItem(myDialog,OPTIONS_USRITEM1,-1,DrawFrame);
	InstallUserItem(myDialog,OPTIONS_USRITEM2,OK,DrawOKBoxRect);

	ShowWindow(myDialog);
	while(!bQuit) {	/* display dialog using our standard filter proc */			
		ModalDialog((ProcPtr)myDialogFilter,&itemHit);
		switch(itemHit) {
			case OK:
			case Cancel:
				bQuit=TRUE;
				break;
			case CONFIRM_BOX:
			case STARTQ_BOX:
			case INCLUDELF_BOX:
			case HYPERLINK_BOX:
				ToggleRadioButton(myDialog,itemHit);
				bChanged = TRUE;
				break;
		}
	}
		
	/* if user clicked OK and something has changed we keep new settings */
	if(itemHit == OK && bChanged) {
		gPrefs.confirmChg = (Boolean)GetRadioButton(myDialog,CONFIRM_BOX);
		gPrefs.startQOpen = (Boolean)GetRadioButton(myDialog,STARTQ_BOX);
		gPrefs.includeLFCopy = (Boolean)GetRadioButton(myDialog,INCLUDELF_BOX);
		gPrefs.markHyperLink = (Boolean)GetRadioButton(myDialog,HYPERLINK_BOX);
		
		ErrorMsg(ERR_PREFCHGWARN);
	}
	
	DisposDialog(myDialog);
}
